from sanic import json
from sanic.log import logger
from ivylink.web.core import IvyLinkGeoIpView

EMPTY_ASN_DATA = {'autonomous_system_number': 0, 'autonomous_system_organization': ''}
EMPTY_LOCATION_DATA = {
    'city': '',
    'state': '',
    'country': '',
    'continent': '',
}


class GeoIpAsn(IvyLinkGeoIpView):
    """
    Get ASN information of IP Address.
    """

    async def post(self, request):

        asn_db = request.app.ctx.geoip_asn_db
        ip = self.load_ip_address(request.json)

        data = get_asn_values(asn_db.get(ip))
        data['ip'] = ip

        return json(data)


class GeoIpLocation(IvyLinkGeoIpView):
    """
    Get Country / City information of IP Address.
    """

    async def post(self, request):

        city_db = request.app.ctx.geoip_city_db
        ip = self.load_ip_address(request.json)

        data = get_location_values(ip, city_db.get(ip))
        data['ip'] = ip

        return json(data)


class GeoIpAll(IvyLinkGeoIpView):
    """
    Get All information of IP Address.
    """

    async def post(self, request):

        asn_db = request.app.ctx.geoip_asn_db
        city_db = request.app.ctx.geoip_city_db
        ip = self.load_ip_address(request.json)

        data = get_asn_values(asn_db.get(ip))
        data |= get_location_values(ip, city_db.get(ip))
        data['ip'] = ip

        return json(data)


def get_asn_values(data: dict):
    """

    :param data:
    :return:
    """
    if not data:
        return EMPTY_ASN_DATA
    return data


def get_location_values(ip: str, data: dict):
    """
    Return fields of interest from city IP data.
    city_ip_data = {
        'city': {'names': {'en': 'Princeton'}},
        'continent': {'code': 'NA', 'geoname_id': 6255149, 'names': {'de': 'Nordamerika', 'en': 'North America', 'es': 'Norteamérica', 'fa': ' امریکای شمالی', 'fr': 'Amérique Du Nord', 'ja': '北アメリカ大陸', 'ko': '북아메리카', 'pt-BR': 'América Do Norte', 'ru': 'Северная Америка', 'zh-CN': '北美洲'}},
        'country': {'geoname_id': 6252001, 'is_in_european_union': False, 'iso_code': 'US', 'names': {'de': 'Vereinigte Staaten von Amerika', 'en': 'United States', 'es': 'Estados Unidos de América (los)', 'fa': 'ایالات متحدهٔ امریکا', 'fr': 'États-Unis', 'ja': 'アメリカ合衆国', 'ko': '미국', 'pt-BR': 'Estados Unidos', 'ru': 'США', 'zh-CN': '美国'}},
        'location': {'latitude': 40.3573, 'longitude': -74.6672},
        'subdivisions': [{'names': {'en': 'New Jersey'}}],
    }
    """
    if not data:
        return EMPTY_LOCATION_DATA

    try:
        return {
            'city': data['city']['names']['en'],
            'state': data['subdivisions'][0]['names']['en'],
            'country': data['country']['iso_code'],
            'continent': data['continent']['names']['en'],
        }
    except KeyError as e:
        logger.info('Unable to retrieve location data for %s. %s', ip, e)
        return EMPTY_LOCATION_DATA


__all__ = (
    'GeoIpAsn',
    'GeoIpLocation',
    'GeoIpAll',
)
